use std::num::NonZero;

pub type Node = NonZero<u32>;

#[derive(Debug)]
pub enum Tag {
    Undefined,
    Root,
    FunctionProto {
        /// Ident
        name: Node,
        /// ParameterList
        parameters: Option<Node>,
        return_type: Node,
    },
    ParameterList {
        /// [Parameter]
        parameters: Vec<Node>,
    },
    Parameter {
        /// Ident
        name: Node,
        /// TypeName
        ty: Node,
    },
    // TypeName: meta-tag of Ident | PrimitiveType | Pointer
    Pointer {
        /// TypeName
        pointee: Node,
    },
    PointerQualifier {
        constness: bool,
    },
    IntegralType(IntegralType),
    PrimitiveType(PrimitiveType),
    FunctionDecl {
        /// FunctionProto
        proto: Node,
        /// Block
        body: Node,
    },
    Ident {
        name: String,
    },
    IntegralConstant {
        bits: u64,
        ty: IntegralType,
    },
    FloatingConstant {
        bits: u64,
        ty: FloatingType,
    },
    Block {
        /// ReturnStmt | ExprStmt | VarDecl
        statements: Vec<Node>,
        trailing_expr: Option<Node>,
    },
    ReturnStmt {
        expr: Option<Node>,
    },
    ExprStmt {
        expr: Node,
    },
    VarDecl {
        let_or_var: LetOrVar,
        /// Ident
        name: Node,
        /// TypeName
        explicit_type: Option<Node>,
        assignment: Option<Node>,
    },
    CallExpr {
        /// Ident | Expr
        lhs: Node,
        /// ArgumentList
        rhs: Option<Node>,
    },
    ArgumentList {
        /// [Argument]
        parameters: Vec<Node>,
    },
    Argument {
        /// Ident
        name: Option<Node>,
        /// expr
        expr: Node,
    },
    ExplicitCast {
        lhs: Node,
        /// TypeName
        typename: Node,
    },
    Deref {
        lhs: Node,
    },
    Ref {
        lhs: Node,
    },
    Not {
        lhs: Node,
    },
    Negate {
        lhs: Node,
    },
    Or {
        lhs: Node,
        rhs: Node,
    },
    And {
        lhs: Node,
        rhs: Node,
    },
    BitOr {
        lhs: Node,
        rhs: Node,
    },
    BitAnd {
        lhs: Node,
        rhs: Node,
    },
    BitXOr {
        lhs: Node,
        rhs: Node,
    },
    Eq {
        lhs: Node,
        rhs: Node,
    },
    NEq {
        lhs: Node,
        rhs: Node,
    },
    Lt {
        lhs: Node,
        rhs: Node,
    },
    Gt {
        lhs: Node,
        rhs: Node,
    },
    Le {
        lhs: Node,
        rhs: Node,
    },
    Ge {
        lhs: Node,
        rhs: Node,
    },
    Shl {
        lhs: Node,
        rhs: Node,
    },
    Shr {
        lhs: Node,
        rhs: Node,
    },
    Add {
        lhs: Node,
        rhs: Node,
    },
    Sub {
        lhs: Node,
        rhs: Node,
    },
    Mul {
        lhs: Node,
        rhs: Node,
    },
    Rem {
        lhs: Node,
        rhs: Node,
    },
    Div {
        lhs: Node,
        rhs: Node,
    },
    Assign {
        lhs: Node,
        rhs: Node,
    },
}

#[derive(Debug)]
pub enum LetOrVar {
    Let,
    Var,
}

#[derive(Debug)]
pub struct IntegralType {
    pub signed: bool,
    pub bits: u16,
}

impl ToString for IntegralType {
    fn to_string(&self) -> String {
        format!("{}{}", if self.signed { "i" } else { "u" }, self.bits)
    }
}

#[derive(Debug)]
pub enum FloatingType {
    Binary32,
    Binary64,
}

impl ToString for FloatingType {
    fn to_string(&self) -> String {
        match self {
            FloatingType::Binary32 => "binary32".to_owned(),
            FloatingType::Binary64 => "binary64".to_owned(),
        }
    }
}

impl IntegralType {
    pub fn u32() -> IntegralType {
        Self {
            signed: false,
            bits: 32,
        }
    }
}

#[derive(Debug)]
pub enum PrimitiveType {
    FloatingType(FloatingType),
    IntegralType(Node),
    Bool,
    Void,
}

impl ToString for PrimitiveType {
    fn to_string(&self) -> String {
        match self {
            PrimitiveType::FloatingType(f) => f.to_string(),
            PrimitiveType::IntegralType(i) => i.to_string(),
            PrimitiveType::Bool => "bool".to_owned(),
            PrimitiveType::Void => "void".to_owned(),
        }
    }
}
